# ---------------------------------------------------------------------------
# Sanitize
# A Plugin for Movable Type
#
# Release 1.2
# October 6, 2002
#
# From Brad Choate
# http://www.bradchoate.com/
# ---------------------------------------------------------------------------
# This software is provided as-is.
# You may use it for commercial or personal use.
# If you distribute it, please keep this notice intact.
#
# Copyright (c) 2002 Brad Choate
# ---------------------------------------------------------------------------
# Release 1.3
# October 30, 2003
#
# Modified to EXPORT_OK sanitize_html
# ---------------------------------------------------------------------------

package bradchoate::sanitize;

use strict;

use vars qw($VERSION @ISA @EXPORT_OK);

use Exporter;

@ISA = qw(Exporter);

$VERSION = '1.03';

@EXPORT_OK = qw( sanitize_html );

# ---------------------------------------------------------------------------

sub sanitize_html {
    my ($s, $a) = @_;

    my (%ok_tags, %global_attrs, %tag_attr);
    foreach (split /,/, $a) {
        my (%ok_attr, $tag, $style);
        $tag = lc($_);
        if ($tag =~ m|^([^ ]+) (.+)$|) {
            my $attrs = $2;
            $tag = lc($1);
            ($style) = $tag =~ m|([/])$|;
            $tag =~ s|$style$|| if $style;
            foreach (split / /, $attrs) {
                $ok_attr{$_} = 1;
            }
            if ($tag eq '*') {
                %global_attrs = %ok_attr;
                next;
            }
        } else {
            ($style) = $tag =~ m|([/])$|;
            $tag =~ s|$style$|| if $style;
        }
        if ($style) {
            $tag_attr{$tag} = $style;
        }
        if (%ok_attr) {
            $ok_tags{$tag} = \%ok_attr;
        } else {
            $ok_tags{$tag} = 1;
        }
    }

    my (@open_tags, %open_tags);
    my $out = '';
    my $start = 0;
    my $last_pos = 0;
    my $seq = 0;
    while ($s =~ m|<|gs) {
        my $start = (pos $s) - 1;
        my $inside = substr($s, $start + 1, 5); # 5 chars should do it...
	if ($last_pos < $start) {
	    $out .= substr($s, $last_pos, $start - $last_pos);
	}

	if ($inside =~ m|^%|) {
	    # asp, jsp-- strip up to %>
	    if ($s =~ m|%>|gs) {
		$last_pos = pos $s;
	    } else {
		# no end in sight-- exit loop
		$last_pos = pos $s;
		last;
	    }
	} elsif ($inside =~ m|^!--|) {
	    pos $s = $start + 4;
	    # html comment or SSI block
	    if ($s =~ m|-->|gs) {
		$last_pos = pos $s;
	    } else {
		# no end in sight-- exit loop
		$last_pos = pos $s;
		last;
	    }
	} elsif ($inside =~ m|^\?|) {
	    # processing instruction or php? strip to ?>
	    if ($s =~ m|\?>|gs) {
		$last_pos = pos $s;
	    } else {
		# can't find the closure.
		$last_pos = pos $s;
		last;
	    }
	} elsif ($inside =~ m|^>|) {
	    $last_pos = $start + 1;
        } elsif ($inside =~ m|^<|) {
	    $last_pos = $start + 1;
	} elsif ($s =~ m|(.+?)>|gs) {
	    $inside = $1;
	    my $end = pos $s;
	    my $name;
	    my $closure = 0;
	    if ($inside =~ m/^([^ ]+) (.+)$/s) {
		$name = lc($1);
                $inside = $2;
	    } else {
		$name = lc($inside);
                $inside = '';
	    }
            if ($name =~ m|^/|) {
                $name =~ s|^/||;
                $closure = 1;
	    }
            if ($inside =~ m|/$|s) {
                $inside =~ s|/$||s;
                $closure = 2;
            }
	    if ($ok_tags{$name} ||
                (exists $tag_attr{$name} && $tag_attr{$name} eq '/')) {
                if ($inside) {
                    my @attrs;
                    while ($inside =~ m/(\w+)\s*=\s*(['"])(.*?)\2/gs) {
                        if ($global_attrs{lc($1)} ||
                            (ref $ok_tags{$name} && $ok_tags{$name}{lc($1)}) ||
                            (ref $ok_tags{"$name/"} && $ok_tags{"$name/"}{lc($1)})) {
                            push @attrs, "$1=$2$3$2";
                        }
                    }
                    if (@attrs) {
                        $inside = join ' ', @attrs;
                    } else {
                        $inside = '';
                    }
                }
                if (exists $tag_attr{$name} && ($tag_attr{$name} eq '/')) {
                    $closure = 2;
                }
                if ($closure != 1 || ($closure == 1 && $open_tags{$name})) {
                    if ($closure == 1) {
                        $out .= _expel_up_to(\@open_tags, \%open_tags, $name);
                    } elsif (!$closure) {
                        ###if (@open_tags &&
                        ###    ($open_tags{$name}) &&
                        ###    (exists $tag_attr{$name}) &&
                        ###    ($tag_attr{$name} eq '!')) {
                        ###    # unnestable tag-- expel stuff
                        ###    $out .= _expel_up_to(\@open_tags, \%open_tags, $name);
                        ###    $out .= '</' . $name . '>';
                        ###}
                        push @open_tags, $name;
                        $open_tags{$name}++; 
                    }
                    $out .= '<'.($closure==1?'/':'').
                        $name.
                        ($inside ? ' '.$inside : '').
                        ($closure==2?' /':'').'>';
                    if ($closure == 1) {
                        $open_tags{$name}--;
                    }
                }
            }
	    $last_pos = $end;
	} else {
	    $last_pos = pos $s;
	    last;
	}
    }
    if (defined $last_pos && ($last_pos < length($s))) {
	$out .= substr($s, $last_pos);
    }
    $out .= '</' . $_ . '>' foreach @open_tags;
    $out;
}

sub _expel_up_to {
    my ($open_tags, $open_hash, $stop_tag) = @_;
    my $out = '';
    while (@$open_tags &&
           (!defined $stop_tag || $open_tags->[$#$open_tags] ne $stop_tag)) {
        my $t = pop @$open_tags;
        $open_hash->{$t}--;
        $out .= '</' . $t . '>';
    }
    if (@$open_tags) {
        my $t = pop @$open_tags;
    }
    $out;
}

1;
